import pytest

from wxc.api import get_obj, get_sourcefile
from wxc.cli import main

TO_CHECK = frozenset(("os.fspath", "math.sqrt"))


@pytest.mark.parametrize("name", TO_CHECK)
def test_get_data_builtin(name):
    obj = get_obj(name)
    with pytest.raises(
        TypeError,
        match=(
            r"module, class, method, function, traceback, frame, or code object was expected,"
            " got builtin_function_or_method"
        ),
    ):
        get_sourcefile(obj)


@pytest.mark.parametrize("name", TO_CHECK)
def test_cli_builtin(name, capsys):
    ret = main([name])
    out, err = capsys.readouterr()
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert (
        err
        == f"ERROR failed to locate source data. {name!r} is a C-compiled function.\n"
    )
    assert ret != 0



import sys
from pathlib import Path

import pytest  # type: ignore

from wxc.api import get_full_data, get_version


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="On Windows, we can't even locate these files",
)
@pytest.mark.parametrize(
    "name,except_python_version", [("math", True), ("platform", False)]
)
def test_stdlib_versions(name: str, except_python_version: bool):
    imp = get_full_data(name)
    assert imp["in_stdlib"]
    assert "version" in imp
    assert except_python_version is imp["version"].startswith("Python")


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="parsing is more convoluted when ':' is a normal path element",
)
def test_finder(package_name):
    pytest.importorskip(package_name)
    imp = get_full_data(package_name)
    assert "source" in imp

    filename, _, line = imp["source"].partition(":")
    p = Path(filename)
    assert p.exists()
    if not imp["in_stdlib"]:
        assert package_name in p.parts


def test_get_obj():
    line1 = get_full_data("pathlib.Path")["source"].partition(":")[-1]
    line2 = get_full_data("pathlib.Path.chmod")["source"].partition(":")[-1]
    assert line1 != ""
    assert line2 != ""
    assert line1 != line2


def test_wrong_package():
    with pytest.raises(ImportError):
        get_full_data("NotARealPackage")


def test_wrong_member():
    with pytest.raises(
        AttributeError,
        match="module 'os.path' has no attribute 'NotARealMember'",
    ):
        get_full_data("os.path.NotARealMember")


def test_get_version_platform():
    # platform.version is a function
    # while get_version is currently able to retrieve the correct result
    # without hitting this problem, this test improves coverage of the defensive
    # programming that was deployed when this special case was discovered
    ret = get_version("platform", _version_attr_lookup_table=("version", "__version__"))
    assert isinstance(ret, str)



def levenshtein_distance(seq1: str, seq2: str, max_dist: int) -> int:
    """
    Compute the levenshtein distance between seq1 and seq2.
    From https://stackabuse.com/levenshtein-distance-and-text-similarity-in-python/

    Parameters
    ----------
    seq1, seq2 : str
        The strings to compute the distance between
    max_dist : integer
        If not None, maximum distance returned (see notes).

    Returns
    -------
    The Levenshtein distance as an integer.

    Notes
    -----
    This computes the Levenshtein distance, i.e. the number of edits to change
    seq1 into seq2. If a maximum distance is passed, the algorithm will stop as soon
    as the number of edits goes above the value. This allows for an earlier break
    and speeds calculations up.
    """
    size_x = len(seq1) + 1
    size_y = len(seq2) + 1
    max_dist = min(max(size_x, size_y), max_dist)

    if abs(size_x - size_y) > max_dist:
        return max_dist + 1

    matrix = []
    for _ in range(size_x):
        # this somewhat clumsy initialization is intended to avoid
        # side effects of the "smart" way. See
        # >>> matrix = [[0]*size_y]*size_x
        # >>> matrix[0] is matrix[1]
        # True
        matrix.append([0] * size_y)
    for x in range(size_x):
        matrix[x][0] = x
    for y in range(size_y):
        matrix[0][y] = y

    for x in range(1, size_x):
        for y in range(1, size_y):
            if seq1[x - 1] == seq2[y - 1]:
                matrix[x][y] = min(
                    matrix[x - 1][y] + 1,
                    matrix[x - 1][y - 1],
                    matrix[x][y - 1] + 1,
                )
            else:
                matrix[x][y] = min(
                    matrix[x - 1][y] + 1,
                    matrix[x - 1][y - 1] + 1,
                    matrix[x][y - 1] + 1,
                )

        # Early break: the minimum distance is already larger than
        # maximum allow value, can return safely.
        if min(matrix[x]) > max_dist:
            return max_dist + 1
    return matrix[size_x - 1][size_y - 1]



import pytest

from wxc.api import get_obj
from wxc.cli import main

BUILTINS_TO_CHECK = frozenset(
    (
        "print",
        "object",
    )
)


@pytest.mark.parametrize("name", BUILTINS_TO_CHECK)
def test_inspect_builtin(name):
    # check that this doesn't fail brutally
    get_obj(name)


@pytest.mark.parametrize("name", BUILTINS_TO_CHECK)
def test_fail_builtins(name, capsys):
    ret = main([name])

    out, err = capsys.readouterr()
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert err == (
        f"ERROR failed to locate source data. {name!r} is a builtin object.\n"
    )
    assert ret != 0



from __future__ import annotations

import builtins
import re
import sys
from collections import defaultdict
from functools import lru_cache
from types import BuiltinFunctionType

from wxc.levenshtein import levenshtein_distance

if False:
    # typecheck only
    from collections.abc import Iterable  # type: ignore [unreachable]
    from typing import Any


# sorted by decreasing order of priority
VERSION_ATTR_LOOKUP_TABLE = ("__version__", "VERSION", "version")


def is_builtin(name: str) -> bool:
    return hasattr(builtins, name)


def is_builtin_func(obj: Any) -> bool:
    if isinstance(obj, str):
        obj = get_obj(obj)
    # this heuristic is equivalent to inspect.isbuiltin()
    # it should be noted that the type 'BuiltinFunctionType' isn't reserved
    # to the standard library, but is characteristic of functions that
    # are "written in C". See
    # https://docs.python.org/3/library/types.html?highlight=builtinfunctiontype#types.BuiltinMethodType
    return isinstance(obj, BuiltinFunctionType)


def get_builtin_obj(name: str):
    return getattr(builtins, name)


def get_suggestions(
    candidates: Iterable[str], target: str, *, max_dist: int = sys.maxsize
) -> list[str]:
    suggestions: dict[int, list[str]] = defaultdict(list)
    minimal_distance: int = max_dist
    for a in candidates:
        d: int = levenshtein_distance(target, a, max_dist=minimal_distance)
        if d <= minimal_distance:
            suggestions[d].append(a)
            minimal_distance = d
    return sorted(suggestions[minimal_distance])


@lru_cache(maxsize=128)
def get_objects(name: str) -> list:
    if is_builtin(name):
        return [get_builtin_obj(name)]

    from importlib import import_module

    name_in = name
    attrs = []
    objects = []
    while name:
        try:
            obj = import_module(name)
            break
        except ImportError:
            name, _, attr = name.rpartition(".")
            attrs.append(attr)
    if not name:
        raise ImportError(name_in)

    objects.append(obj)

    for attr in reversed(attrs):
        try:
            obj = getattr(obj, attr)
        except AttributeError as exc:
            msg = exc.args[0]
            # force the name to match the one specified by the user even
            # in cases where they are using an alias (for instance os.path is a alias for posixpath on UNIX)
            msg = re.sub(r"\'[^-\s]*\'", f"{name!r}", msg, count=1)
            suggestions = get_suggestions(dir(obj), attr)
            if len(suggestions) > 1:
                repr_suggestions = ", ".join(f"{s!r}" for s in suggestions)
                msg += f". Here are the closest matches: {repr_suggestions}"
            elif len(suggestions) == 1:
                msg += f". Did you mean {suggestions[0]!r} ?"
            raise AttributeError(msg) from exc
        else:
            name += f".{attr}"
            objects.append(obj)

    return objects


def get_obj(name: str):
    # for backwards compatibility
    return get_objects(name)[-1]


def get_sourcefile(obj):
    import inspect

    try:
        file = inspect.getfile(obj)
    except OSError:
        file = obj.__file__
    except TypeError:
        # this happens for instance with `math.sqrt`
        # because inspect.getfile doesn't work on compiled code
        # the second condition is met for os.fspath
        if inspect.ismodule(obj) or is_builtin_func(obj):
            raise
        if isinstance(obj, property):
            raise
        return get_sourcefile(inspect.getmodule(obj))
    return file


def get_sourceline(obj):
    import inspect

    return inspect.getsourcelines(obj)[1]


def get_version(
    package_name: str, *, _version_attr_lookup_table=VERSION_ATTR_LOOKUP_TABLE
) -> str:
    package = get_obj(package_name)
    for version_attr in _version_attr_lookup_table:
        if hasattr(package, version_attr):
            retv = getattr(package, version_attr)
            if not isinstance(retv, str):
                # this conditional guards against rare cases like the builtin
                # platform module, platform.version being a function
                continue
            return retv

    import importlib.metadata as md

    try:
        return str(md.version(package_name))
    except md.PackageNotFoundError:
        pass

    if package_name in sys.stdlib_module_names:
        from platform import python_version

        return f"Python {python_version()}"

    raise LookupError(f"Could not determine version metadata from {package_name!r}")


def get_full_data(name: str) -> dict:
    data = defaultdict(str)
    package_name, _, _ = name.partition(".")

    objects = get_objects(name)

    for obj in reversed(objects):
        try:
            source = get_sourcefile(obj)
        except RecursionError:
            pass
        except TypeError:
            # as of Python 3.11, inspect.getfile doesn't have support for properties
            # but we're not making this a hard failure in case it is added in the future
            # and we fallback to finding out the sourcefile of the class itself
            if isinstance(obj, property):
                continue
            else:
                raise
        else:
            try:
                lineno = get_sourceline(obj)
            except (OSError, TypeError):
                pass
            else:
                source += f":{lineno}" if lineno else ""
                break
            finally:
                data["source"] = source

    try:
        data["version"] = get_version(package_name)
    except LookupError:
        pass

    data["in_stdlib"] = package_name in sys.stdlib_module_names

    return data



import sys
from importlib import import_module
from importlib.util import find_spec

import pytest

from wxc.cli import main

valid_queries = [None, "--version", "-v", "--full", "-f"]


@pytest.mark.parametrize("arg", valid_queries)
def test_query_valid_field(arg):
    argv = ["os.path.join"]
    if arg is not None:
        argv.append(arg)
    main(argv)


def test_elementary_queries(capsys, package_name):
    try:
        import_module(package_name)
    except ImportError:
        pytest.skip()

    ret = main([package_name, "--version"])

    out, err = capsys.readouterr()

    if package_name == "math" and sys.platform.startswith("win"):
        # rich may output an unspecified amount of newlines
        # that don't actually affect the result visually
        assert out.strip() == ""
        assert err == "ERROR failed to locate source data.\n"
        assert ret != 0
        return

    assert out != "unknown"
    assert err == ""
    assert ret == 0


@pytest.mark.parametrize("arg", valid_queries)
def test_falty_queries(capsys, arg):
    argv = ["NotARealPackage"]
    if arg is not None:
        argv.append(arg)
    res = main(argv)
    assert res != 0
    out, err = capsys.readouterr()
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert err == "ERROR no installed package with name 'NotARealPackage'\n"


def test_stdlib_typos_in_module_name(capsys):
    res = main(["sis"])
    assert res != 0
    out, err = capsys.readouterr()
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert err == "ERROR no installed package with name 'sis'. Did you mean 'sys' ?\n"


def test_non_existing_member(capsys):
    ret = main(["pathlib.nothing"])
    assert ret != 0
    out, err = capsys.readouterr()
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""

    assert err.startswith(
        "ERROR module 'pathlib' has no attribute 'nothing'."
        " Here are the closest matches:"
    )


def test_compiled_source(capsys):
    pytest.importorskip("numpy")
    ret = main(["numpy.abs.at"])
    out, err = capsys.readouterr()
    assert (
        err
        == "ERROR failed to locate source data. 'numpy.abs.at' is a C-compiled function.\n"
    )
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert ret != 0


def test_typo1(capsys):
    ret = main(["pathlib.Path.chmode"])
    out, err = capsys.readouterr()
    assert (
        err
        == "ERROR type object 'pathlib.Path' has no attribute 'chmode'. Did you mean 'chmod' ?\n"
    )
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert ret != 0


def test_typo2(capsys):
    ret = main(["pathlib.Path.homme"])
    out, err = capsys.readouterr()
    assert (
        err
        == "ERROR type object 'pathlib.Path' has no attribute 'homme'. Did you mean 'home' ?\n"
    )
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert ret != 0


@pytest.mark.skipif(
    find_spec("matplotlib_inline") is None,
    reason="matplotlib_inline isn't a hard dependency",
)
def test_normalize_hyphen(capsys):
    # note that use a simple pytest.mark.skipif instead of pytest.importorskip
    # because the latter doesn't seem to play nice with coverage
    ret = main(["matplotlib-inline"])
    out, err = capsys.readouterr()
    assert err == ""
    assert out
    assert ret == 0


def test_source(fake_module, capsys):
    expected = (
        """\n"""
        """  4 def import_me_if_you_can():\n"""
        '''  5     """Docstrings are good. Use them."""\n'''
        """  6     return "Gotcha"\n"""
        """  7 \n"""
    )
    syspath, name = fake_module
    ret = main([f"{name}.import_me_if_you_can", "--source", "--lines"])
    out, err = capsys.readouterr()
    assert out.startswith(expected)
    assert err == ""
    assert ret == 0


def test_source_error(capsys):
    pytest.importorskip("numpy")
    ret1 = main(["numpy.ndarray"])
    assert ret1 == 0
    out, err = capsys.readouterr()
    assert out != ""
    assert err == ""

    ret2 = main(["numpy.ndarray", "-s"])
    out, err = capsys.readouterr()
    # rich may output an unspecified amount of newlines
    # that don't actually affect the result visually
    assert out.strip() == ""
    assert err == "ERROR could not find class definition\n"
    assert ret2 != 0


def test_lines_no_source(capsys):
    ret1 = main(["inspect.getfile", "--lines"])
    assert ret1 != 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == "ERROR --lines flag is meaningless if --source isn't passed too\n"


def test_property(capsys):
    ret = main(["rich.progress.Task.started"])
    assert ret == 0
    out, err = capsys.readouterr()
    assert out != ""
    assert err == ""



import sys
from pathlib import Path

import pytest
from schema import Optional, Schema

from wxc.api import get_full_data

template = Schema(
    {
        "source": str,
        "in_stdlib": bool,
        Optional("version"): str,
    }
)


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="parsing is more convoluted when ':' is a normal path element",
)
def test_empty_module_query(fake_module):
    """Check for robustness of get_full_data
    with an empty module (in particular, no version data)
    """
    syspath, name = fake_module
    data = get_full_data(name)
    filename, _, _ = data["source"].partition(":")
    assert Path(syspath, name) in Path(filename).parents
    assert "version" not in data

    template.validate(data)


def test_field_member():
    d1 = get_full_data("os.path")
    d2 = get_full_data("os.path.expanduser")

    template.validate(d1)
    template.validate(d2)

    assert d2["version"] == d1["version"]
    assert d2["in_stdlib"] is True
    assert d1["in_stdlib"] is True



import sys
from argparse import ArgumentParser

from wxc.api import (
    get_full_data,
    get_obj,
    get_sourceline,
    get_suggestions,
    is_builtin,
    is_builtin_func,
)

builtin_print = print


def print_err(msg):
    from rich import print

    print(f"[bold white on red]ERROR[/] {msg}", file=sys.stderr)


class ScopeName(str):
    def __new__(cls, content):
        return str.__new__(cls, content.replace("-", "_"))


def main(argv: list[str] | None = None) -> int:
    parser = ArgumentParser()
    parser.add_argument(
        "name",
        type=ScopeName,
        help="target Python scope (package.module.submodule.class.method)",
    )
    command_group = parser.add_mutually_exclusive_group()
    command_group.add_argument(
        "-v", "--version", action="store_true", help="print module version"
    )
    command_group.add_argument(
        "-f",
        "--full",
        action="store_true",
        help="print a full report",
    )
    parser.add_argument(
        "-s", "--source", action="store_true", help="print the source code"
    )
    parser.add_argument(
        "--lines",
        dest="show_lines",
        action="store_true",
        help="show source lines",
    )
    args = parser.parse_args(argv)

    if args.show_lines and not args.source:
        print_err("--lines flag is meaningless if --source isn't passed too")
        return 1

    if "." not in args.name:
        # this is a simple module request
        # let's try to get the result without actually importing it first
        if args.version:
            import importlib.metadata as md

            try:
                version = md.version(args.name)
            except md.PackageNotFoundError:
                pass  # resort to expensive search
            else:
                builtin_print(version)
                return 0
        elif args.full:
            # quick lookup not possible (or at least, not implemented)
            pass
        else:
            from importlib.util import find_spec

            spec = find_spec(args.name)
            if spec is None:
                pass  # resort to expensive search
            else:
                from rich import print

                print(spec.origin)
                return 0

    from rich.progress import Progress, SpinnerColumn, TextColumn

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task("[bold]Resolving source")
        try:
            obj = get_obj(args.name)
        except ImportError:
            package_name = args.name.partition(".")[0]
            msg = f"no installed package with name {package_name!r}"
            suggestions = get_suggestions(
                sys.builtin_module_names, package_name, max_dist=2
            )
            if len(suggestions) == 1:
                msg += f". Did you mean {suggestions[0]!r} ?"
            progress.stop()
            print_err(msg)
            return 1
        except AttributeError as exc:
            progress.stop()
            print_err(exc)
            return 1

    try:
        data = get_full_data(args.name)
    except TypeError:
        msg = "failed to locate source data."
        if is_builtin(args.name):
            msg += f" {args.name!r} is a builtin object."
        elif is_builtin_func(args.name):
            msg += f" {args.name!r} is a C-compiled function."
        print_err(msg)
        return 1

    if args.full:
        from rich import print

        data["name"] = args.name
        ver = f"version = {data.pop('version', 'unknown')}"
        print("\n".join(f"{k} = {v}" for k, v in data.items()))
        builtin_print(ver)
        return 0

    if args.version:
        if not data["version"]:
            print_err(f"did not find version metadata for {args.name!r}")
            return 1
        builtin_print(data["version"])
        return 0

    if "source" not in data:
        print_err(f"did not resolve source file for {args.name!r}")
        return 1

    if args.source:
        import inspect

        from rich.syntax import Syntax

        try:
            code = inspect.getsource(obj)
        except OSError as exc:
            # inspect.getsource _can_ be the first failing call so we wrap its
            # error because it's in line with wxc's own error messages and
            # there's probably not much else we can do about it.
            print_err(exc)
            return 1

        from rich import print

        print(
            Syntax(
                code,
                "python",
                theme="monokai",
                line_numbers=args.show_lines,
                start_line=get_sourceline(obj),
                background_color="default",
            )
        )

    from rich import print

    print(data["source"])

    return 0
